#ifndef HELPER_H
#define HELPER_H

#include "capd.h"

/**
 * helper function for convenient creation of Rect2Sets.
 * @param x double center of set x coordinate
 * @param y double center of set y coordinate
 * @param w double width of set
 * @param h double height of set
 * @param C IntervalMatrix basic transformation such that set = (x,y) + C([-w/2, w/2], [-h/2,h/2])
 */
Rect2Set makeSet(interval x, interval y, double a, double b, IntervalMatrix C){

	interval data2[] = {x, y};
	IntervalVector p(2, data2);
	interval radiusW(-a, a);
	interval radiusH(-b, b);	
	interval data[] = {radiusW, radiusH};
	IntervalVector r(2, data); 
	
	Rect2Set set(p, C, r);
	return set;
}

/**
 * helper function for convenient creation of C1Rect2Sets - sets that can handle C1 computations (C1-Lohner Algorithm).
 * @param x double center of set x coordinate
 * @param y double center of set y coordinate
 * @param w double width of set
 * @param h double height of set
 * @param C IntervalMatrix basic transformation such that set = (x,y) + C([-w/2, w/2], [-h/2,h/2])
 */
C1Rect2Set makeC1Set(interval x, interval y, double a, double b, IntervalMatrix C){

	interval data2[] = {x, y};
	IntervalVector p(2, data2);
	interval radiusW(-a, a);
	interval radiusH(-b, b);	
	interval data[] = {radiusW, radiusH};
	IntervalVector r(2, data); 
	
	C1Rect2Set set(p, C, r, EuclLNorm());
	return set;
}

/**
 * helper function for convenient cretion of Recr2Sets.
 * @param x double center of set x coordinate
 * @param y double center of set y coordinate
 * @param ksi double center value of range of parameter ksi
 * @param w double width of set
 * @param h double height of set
 * @param d double size of parameter range (KSI)
 * @param C IntervalMatrix basic transformation such that set = (x,y) + C([-w/2, w/2], [-h/2,h/2])
 */
C1Rect2SetParam makeSetParam(interval x, interval y, interval ksi, double a, double b, IntervalMatrix C){

	interval data3[] = {x, y, ksi.mid()};
	IntervalVectorParam p(3, data3);
	interval radiusW(-a, a);
	interval radiusH(-b, b);	
	interval data[] = {radiusW, radiusH, (ksi - ksi.mid())};
	IntervalVectorParam r(3, data); 
	
	//C matrix is 2x2, CC matrix is 3x3 (extended system)
	IntervalMatrixParam CC = IntervalMatrixParam::Identity(3);
	CC(1,1) = C(1,1);
	CC(2,1) = C(2,1);
	CC(1,2) = C(1,2);
	CC(2,2) = C(2,2);	
	
	C1Rect2SetParam set(p, CC, r, EuclLNormParam());
	return set;
}

/**
 * helper function, creates 2 dimensional interval vector with [x,x],[y,y] entries
 */
IntervalVector makeVector(double x, double y){
	interval data[] = {x, y};
	return IntervalVector(DIMENSION, data);
}
/**
 * helper function, creates 2 dimensional interval vector with x, y entries
 */
IntervalVector makeVector(interval x, interval y){
	interval data[] = {x, y};
	return IntervalVector(DIMENSION, data);
}

/**
 * Simple inverse of 2x2 matrix!
 */
IntervalMatrix inverse2x2(const IntervalMatrix &C){
	IntervalMatrix IC = IntervalMatrix::Identity(2);
	IC(1,1) =  C(2,2);	
	IC(2,2) =  C(1,1);
	IC(1,2) = -C(1,2);
	IC(2,1) = -C(2,1);
	IC /= (IC(1,1)*IC(2,2)) - (IC(1,2)*IC(2,1));
	return IC;
}

/**
 * we need it for intervals etc.
 */
template<typename T>
T abs (const T& a){
	return a < 0 ? -a : a;
}

/**
 * Returns matrix such that P*M*P^-1 = Id (in interval sense...)
 */
IntervalMatrix diagonalize(IntervalMatrix &M){
	double a,b,c,d;
	a = M(1,1).mid().leftBound();
	b = M(1,2).mid().leftBound();
	c = M(2,1).mid().leftBound();
	d = M(2,2).mid().leftBound();
	double delta = (a+d)*(a+d)-4*a*d+4*c*b;
	double l1 = 0.5*((a+d)-sqrt(delta));
	double l2 = 0.5*((a+d)+sqrt(delta));
	if  (l1 < l2){
		double tmp = l2;
		l2 = l1;
		l1 = tmp;
	}
	
	double y1 = sqrt((c*c)/(c*c+(l1-d)*(l1-d)));
	double x1 = sqrt(1-y1*y1);
	x1 = -x1; //trick
	
	double y2 = sqrt((c*c)/(c*c+(l2-d)*(l2-d)));
	double x2 = sqrt(1-y2*y2);	
	x2 = -x2; //trick			
	
	if (std::abs(a*x1+b*y1-l1*x1) > std::abs(-a*x1+b*y1+l1*x1)){ ///trick to automaticaly choose good coordinates
		x1 = -x1;
	}
	
	if (std::abs(a*x2+b*y2-l2*x2) > std::abs(-a*x2+b*y2+l2*x2)){ ///trick to automaticaly choose good coordinates
		x2 = -x2;
	}
				
	IntervalMatrix P = IntervalMatrix::Identity(2);
	P(1,1) = x1;
	P(1,2) = x2;
	P(2,1) = y1;
	P(2,2) = y2;
	
	return P;
}

/**
 * Returns matrix such that P*M*P^-1 = Id (in interval sense...)
 */
IntervalMatrix i_diagonalize(IntervalMatrix &M){
	interval a,b,c,d;
	a = M(1,1);
	b = M(1,2);
	c = M(2,1);
	d = M(2,2);
	interval delta = power(a+d, 2)-4*a*d+4*c*b;
	interval l1 = 0.5*((a+d)-sqrt(delta));
	interval l2 = 0.5*((a+d)+sqrt(delta));
	if  (l1 < l2){
		interval tmp = l2;
		l2 = l1;
		l1 = tmp;
	}
	
	interval y1 = sqrt(power(c, 2)/(power(c,2)+power(l1-d, 2)));
	interval x1 = sqrt(1-power(y1, 2));
	x1 = -x1; //trick
	
	interval y2 = sqrt(power(c, 2)/(power(c, 2)+power(l2-d, 2)));
	interval x2 = sqrt(1-power(y2, 2));	
	x2 = -x2; //trick			
	
	if (abs(a*x1+b*y1-l1*x1) > abs(-a*x1+b*y1+l1*x1)){ ///trick to automaticaly choose good coordinates
		x1 = -x1;
	}
	
	if (abs(a*x2+b*y2-l2*x2) > abs(-a*x2+b*y2+l2*x2)){ ///trick to automaticaly choose good coordinates
		x2 = -x2;
	}
				
	IntervalMatrix P = IntervalMatrix::Identity(2);
	P(1,1) = x1;
	P(1,2) = x2;
	P(2,1) = y1;
	P(2,2) = y2;
	
	return P;
}

namespace ConeCondSet{

/** creates set containing the boundary of h-set that is crossed by stable manifold, and lies on the left side of line of singular points */
static Rect2Set makeStableSet(double x, double y, const IntervalMatrix& P, double a, double b, interval mid){
	IntervalVector rsp = makeVector(interval(-a, a), interval(b));
	rsp = P * rsp;				
	if (rsp[0].leftBound() + x > mid.rightBound()){ 
		//rsp[0] = x > mid - we are on the bad side of the mid - choose other side
		rsp = makeVector(interval(-a, a), interval(-b));
		rsp = P * rsp;				
	}
	Rect2Set RS = makeSet(rsp[0].mid() + x, rsp[1].mid() + y, a, 0.0, P);			
	return RS;
}

/** creates set containing the boundary of h-set that is crossed by unstable manifold, and lies on the left side of line of singular points */
static Rect2Set makeUnstableSet(double x, double y, const IntervalMatrix& P, double a, double b, interval mid){
	IntervalVector rup = makeVector(interval(-a), interval(-b, b));
	rup = P * rup;				
	if (rup[0].leftBound() + x > mid.rightBound()){ 
		//rsp[0] = x > mid - we are on the bad side of the mid - choose other side
		rup = makeVector(interval(a), interval(-b, b));	
		rup = P * rup;				
	}
	Rect2Set RU = makeSet(rup[0].mid() + x, rup[1].mid() + y, 0.0, b, P); 
	return RU;
}

/** creates set containing the boundary of h-set that is crossed by stable manifold, and lies on the left side of line of singular points */
static C1Rect2SetParam makeStableSetParam(double x, double y, const IntervalMatrix& P, double a, double b, double loKSI, double upKSI, interval mid){
	IntervalVector rsp = makeVector(interval(-a, a), interval(b));
	rsp = P * rsp;				
	if (rsp[0].leftBound() + x > mid.rightBound()){
		rsp = makeVector(interval(-a, a), interval(-b));
		rsp = P * rsp;				
	}
	interval ksi(loKSI, upKSI);
	C1Rect2SetParam RS = makeSetParam(rsp[0].mid() + x, rsp[1].mid() + y, ksi, a, 0.0, P);					
	return RS;
}	

/** creates set containing the boundary of h-set that is crossed by unstable manifold, and lies on the left side of line of singular points */	
static C1Rect2SetParam makeUnstableSetParam(double x, double y, const IntervalMatrix& P, double a, double b, double loKSI, double upKSI, interval mid){
	IntervalVector rup = makeVector(interval(-a), interval(-b, b));
	rup = P * rup;				
	if (rup[0].leftBound() + x > mid.rightBound()){		
		rup = makeVector(interval(a), interval(-b, b));	
		rup = P * rup;				
	}
	interval ksi(loKSI, upKSI);
	C1Rect2SetParam RU = makeSetParam(rup[0].mid() + x, rup[1].mid() + y, ksi, 0.0, b, P); 
	return RU;
}			
	
} //namespace ConeCondSet

/**
 * helper function - inner product of two Vectors in 2D
 */
interval dot(IntervalVector v, IntervalVector n){
	return v[0] * n[0] + v[1] * n[1];
}

#endif
