#ifndef ODE_SYSTEM_H
#define ODE_SYSTEM_H

#include "helper.h"
#include "capd.h"

/**
 * class representing system under consideration
 */
struct ODESystem{
public:
	/** phisical parameters, as in equation */
	interval KAPPA;	// default = 1;
	interval TAU;	// default = 0.05;
	interval GAMMA;	// default = -1;
	interval CHI;	// default = 4;
	/** other dependent parameters, can be computetd */
	/** sigma is defined in work of vladimirov as 1 + TAU * KSI */
	interval SIGMA;	// = 1 + TAU * KSI;
	/** first singular point */
	interval X1, Y1;
	/** second singular point */
	interval X2, Y2;
	/** most interested parameter, for which we will be investigating existance of homoclinic orbit */
	interval KSI;
	
	/** helper function */
	IntervalMatrix getDerivAt(interval x, interval y){
		ODESystem &system = (*this); //trick to do fast refactoring ;)
	
		Function deriv[2][2];
		system.getDeriv(deriv);
		
		Function& df1dx = deriv[0][0];
		Function& df1dy = deriv[0][1];
		Function& df2dx = deriv[1][0];
		Function& df2dy = deriv[1][1];
		
		system.setTranslateParams(df1dx, x, y);
		system.setTranslateParams(df1dy, x, y);
		system.setTranslateParams(df2dx, x, y);
		system.setTranslateParams(df2dy, x, y);
		
		interval x0(0);
		interval y0(0);
		
		system.setTransformParams(df1dx, IntervalMatrix::Identity(2));
		system.setTransformParams(df1dy, IntervalMatrix::Identity(2));
		system.setTransformParams(df2dx, IntervalMatrix::Identity(2));
		system.setTransformParams(df2dy, IntervalMatrix::Identity(2));
		
		df1dx.setParameter("y", y0);
		df1dy.setParameter("y", y0);
		df2dx.setParameter("y", y0);
		df2dy.setParameter("y", y0);	
		
		IntervalMatrix Df00 = IntervalMatrix::Identity(2);
		Df00(1,1) = df1dx(x0);
		Df00(1,2) = df1dy(x0);
		Df00(2,1) = df2dx(x0);
		Df00(2,2) = df2dy(x0);
		
		return Df00;
	}
	
	/** computes other, dependent parameters */
	void computeParams(){
		this->SIGMA = 1 + TAU * KSI;
		this->X1 = -KAPPA / GAMMA;
		this->Y1 = -GAMMA;
		this->X2 = sqrt(CHI / (TAU * power(KSI, 2)));
		this->Y2 = (KAPPA - TAU * KSI * GAMMA * this->X2) / (SIGMA * this->X2);
	}	
	
	/** constructor, creates new instance with apropriate values of basic parameters */
	ODESystem(interval KAPPA = 1,
					  interval TAU   = 0.05,
					  interval GAMMA = -1,
					  interval CHI   = 4){
		this->KAPPA = KAPPA;
		this->TAU   = TAU;
		this->GAMMA = GAMMA;
		this->CHI   = CHI;
		
		this->KSI = -1; //arbitrary for this moment, KSI should be set with system->setKsi(value) before doing any computation.
		
		computeParams();
	}
	
	/** computes $\xi_{cr}$, as in [VV] */
	double getCriticalKSI(){
		return (-((CHI + sqrt(CHI * CHI + 4 * KAPPA * KAPPA * KAPPA / (GAMMA * GAMMA))) / (2 * KAPPA * KAPPA / (GAMMA * GAMMA)))).mid().leftBound();
	}
	
	/** set actual value of parameter $\xi$ */
	void setKSI(interval KSI){
		this->KSI = KSI;
		computeParams();
	}
	
	/** returns the vector containing bounds of singular point $(R_1,\Pi_1)$ */
	IntervalVector getSingular1(){
		interval data[] = {this->X1, this->Y1};
		return IntervalVector(DIMENSION, data);
	}
	/** returns the vector containing bounds of singular point $(R_2,\Pi_2)$ */
	IntervalVector getSingular2(){
		interval data[] = {this->X2, this->Y2};
		return IntervalVector(DIMENSION, data);
	}
	
	/** 
	 * gets function f and set apropriate parameters, KSI, KAPPA, etc. 
	 * it set the parameters:
	 *   s <- KSI ($\xi$)
	 *   o <- SIGMA ($\sigma$)
	 *   k <- KAPPA ($\kappa$)
	 *   u <- TAU ($\tau$)
	 *   q <- GAMMA ($\gamma$)
	 *   w <- CHI ($\chi$)
	 * so function f should define those parameters: s, o, k, u, q, w
	 */
	template<typename FT>
	void setParams(FT &f){
		f.setParameter("o", SIGMA);
		f.setParameter("k", KAPPA);
		f.setParameter("u", TAU);
		f.setParameter("q", GAMMA);
		f.setParameter("w", CHI);
	}
	
	template<typename FT>
	void fixKSI(FT &f){
		f.setParameter("s", KSI);
	}
	
	/** 
	 * implementation of translation of parameters
	 * Map version
	 */
	template<typename FT>
	void setTranslateParams(FT &f, interval x, interval y){
		f.setParameter("r", x);
		f.setParameter("p", y);
	}	
	
	/** 
	 * implementation of C transform of equation
	 * we will use as C transform matrix that "diagonalize" DF(0,0)
	 * map version
	 */
	template<typename FT> 
	void setTransformParams(FT &f, IntervalMatrix C){
		f.setParameter("a", C(1,1));
		f.setParameter("b", C(1,2));
		f.setParameter("c", C(2,1));
		f.setParameter("d", C(2,2));
	}
	
	
	/**
	 * returns map that defines right side of ODE under consideration.
	 * The map may be parametrized with some parameters:
	 * s = KSI
	 * o = SIGMA
	 * k = KAPPA
	 * u = TAU
	 * q = GAMMA
	 * w = CHI
	 * The above parameters are automaticly set to those remembered by ODESystem
	 *
	 * Below are parameters that will be used to change coodinates and so on
	 * p = translation of X, i.e. \hat{X} = X + p
	 * r = translation of Y, i.e. \hat{Y} = Y + r
	 * a,b,c,d = entries of coordinate change matrix, i. e. 
	 *
	 * | a  b |
	 * |      |   = C     (matrix)
	 * | c  d |   
	 * Get map will be used by Proof class
	 */
	Map getMap(){				
		Map f = "var:x,y,s,o,k,u,q,w,p,r,a,b,c,d;fun:-k*(a*x+b*y+r)+o*((a*x+b*y+r)^2)*(c*x+d*y+p)+u*s*q*((a*x+b*y+r)^2),(s^2)*k*(a*x+b*y+r)-s*w*(c*x+d*y+p)-(s^2)*((a*x+b*y+r)^2)*(c*x+d*y+p)-s*w*q;"; 		
		fixKSI(f);
		f.setParameter("o", SIGMA);
		f.setParameter("k", KAPPA);
		f.setParameter("u", TAU);
		f.setParameter("q", GAMMA);
		f.setParameter("w", CHI);		
		return f;
	}
	
	
	MapParam getMapParam(){				
		MapParam f = "var:x,y,z,o,k,u,q,w,p,r,a,b,c,d;fun:-k*(a*x+b*y+r)+o*((a*x+b*y+r)^2)*(c*x+d*y+p)+u*z*q*((a*x+b*y+r)^2),(z^2)*k*(a*x+b*y+r)-z*w*(c*x+d*y+p)-(z^2)*((a*x+b*y+r)^2)*(c*x+d*y+p)-z*w*q,0;"; 		
		f.setParameter("o", SIGMA);
		f.setParameter("k", KAPPA);
		f.setParameter("u", TAU);
		f.setParameter("q", GAMMA);
		f.setParameter("w", CHI);		
		return f;
	}	
	/**
	 * Returns jacobi matrix of right side of equation
	 * jacobi matrix is 2x2 matrix containing functions of x and y as entries
	 */
	void getDeriv(Function deriv[][2]){
		deriv[0][0] = "var:x,y,s,o,k,u,q,w,a,b,c,d,p,r;fun:-k+2*o*(a*x+b*y+r)*(c*x+d*y+p)+2*u*s*q*(a*x+b*y+r);";
		setParams(deriv[0][0]);
		fixKSI(deriv[0][0]);
		deriv[0][1] = "var:x,y,s,o,k,u,q,w,a,b,c,d,p,r;fun:o*((a*x+b*y+r)^2);";
		setParams(deriv[0][1]);
		fixKSI(deriv[0][1]);
		deriv[1][0] = "var:x,y,s,o,k,u,q,w,a,b,c,d,p,r;fun:(s^2)*k-2*(s^2)*(a*x+b*y+r)*(c*x+d*y+p);";
		setParams(deriv[1][0]);
		fixKSI(deriv[1][0]);
		deriv[1][1] = "var:x,y,s,o,k,u,q,w,a,b,c,d,p,r;fun:-s*w-(s^2)*((a*x+b*y+r)^2);";
		setParams(deriv[1][1]);	
		fixKSI(deriv[1][1]);
	}
	/**
	 * returns derivatives by KSI variable
	 */
	void getDerivParam(Function deriv[2]){
		deriv[0] = "var:x,y,s,o,k,u,q,w,a,b,c,d,p,r;fun:u*q*((a*x+b*y+r)^2);";
		setParams(deriv[0]);	
		fixKSI(deriv[0]);
		deriv[1] = "var:x,y,s,o,k,u,q,w,a,b,c,d,p,r;fun:2*s*k*(a*x+b*y+r)-w*(c*x+d*y+p)-2*s*((a*x+b*y+r)^2)*(c*x+d*y+p)-w*q;";
		setParams(deriv[1]);	
		fixKSI(deriv[1]);
	}

};
/* class System *******************************************************************/

#endif
