/*****************************************************************************
 * Robert Szczelina
 * Institute of Computer Science
 * Jagiellonian University
 *
 * robert.szczelina@uj.edu.pl
 *
 * This program can prove existance of homoclinic orbit in some dynamical 
 * system of the form:
 *                          $ \Delta_{\xi}(x)x'=f_{\xi}(x) $
 *
 * near the stationary point lying on the line of singular points \Delta_{\xi}(x)=0.
 * Details of the program are described in [ZS].
 *
 *
 * Compilation instructions are provided in the readme.txt file in the main directory.
 *
 * This program uses a part of the CAPD library 
 * thereby it is distributed under the terms of the GNU General Public License.
 * Consult  http://capd.ii.uj.edu.pl/ for details. 
 *
 * Bibliography:
 *  [ZS] Zgliczy\'nski P., Szczelina R.  'Homoclinic bifurcation in a planar 
 *       singular ODE - a computer assisted proof.', 
 *       http://www.ii.uj.edu.pl/~szczelir/p/papers
 *****************************************************************************/
 
#include <cmath>
#include <stdexcept>
#include <fstream>
#include <iostream>

/** include configuration file */
#include "setup.h"
/** include file that contains output classes */
#include "output.h"
/** include all necessary CAPD components */
#include "capd.h"
/** some help functions */
#include "helper.h"
/** include system of odes to solve... */
#include "ode_system.h"
/** include parameters for the problem */
#include "params.h"

#ifdef DEBUG
typedef CerrOutput DebugOutput;
#else
typedef NullOutput DebugOutput;
#endif

/** standard... */
using namespace std;

/**
 * Class containing all verification steps for our proof of existance of homoclinic orbit
 */
template<typename InfoHandler, typename DebugHandler, typename GraphicsHandler> 
struct Proof{
private:
	/** system under consideration */
	ODESystem system;
	/** a priori lower bound for KSI */
	double loKSI;
	/** a priori upper bound for KSI */
	double upKSI;
	
	/** definition of stable set - x coordinate */
	double a_S;
	/** definition of stable set - y coordinate */
	double b_S; 
	/** definition of unstable set - x coordinate */
	double a_U; 
	/** definition of unstable set - y coordinate */
	double b_U;
	/** section wil be defined as line (x,y) : y = -section */
	double section;
	/** x coordinate of the origin of computations - z_0 */
	double z0_x;
	/** y coordinate of the origin of computations - z_0 */
	double z0_y;
	/** this is x coordinate defining line of singularities Delta for all posiible xi*/
	interval forbiddenX;
	/** this is the interval matrix that contains coordinate changes to space in eigenvector coordinates for Df for all possible values of xi and (R_2, Pi_2) */
	IntervalMatrix P;
	/** contains inverses of the matrix P */
	IntervalMatrix IP;		
	/** this is the interval matrix that contains coordinate changes to space in eigenvector coordinates for Df */
	IntervalMatrix CC;
	/** inverse of the matrix CC */
	IntervalMatrix ICC;	
	/** coordinate matrix M for Unstable set */
	IntervalMatrix M_U;
	/** coordinate matrix M for Stable set */
	IntervalMatrix M_S;

	/** we will use some output messages and we should have way to handle it in different ways */	
	/** to this handler proof will send important information like successful/failure of steps */
	InfoHandler info;
	/** to this handler proof will send more detailed informaton on computation, we can use it to observe process of computing */
	DebugHandler debug;
	/** to this handler proof will send graphics */
	GraphicsHandler graphics;
	
	/** 
	 * Returns Cone Condition Matrix (Df^TQ + QDf) 
	 * - for which we tests positive definitenes. 
	 * Purpose of function: I use it everywhere, DRY 
	 */
	IntervalMatrix coneCondMatrix(IntervalMatrix M, IntervalMatrix Q, double a, double b);
	/** 
	 * verifies cone conditions (theorem 2) for given h-set, returns true if passed or false if failed 
	 */
	bool testConeCond(IntervalMatrix M, double a, double b);
	/**
	 * rigorous integration of sets S and U, returns
	 */
	void integration(Rect2Set& S, Rect2Set& U, IntervalVector& outS, IntervalVector& outU, double ksi, double step, int order);	
	/**
	 * returns L constant from the theorem about parameter estimates	 
	 */
	interval getL(IntervalMatrix M, double a, double b);
	/**
	 * returns A constant from the theorem about parameter estimates	 
	 */	
	interval getA(IntervalMatrix M, IntervalMatrix Q, double a, double b);	
public:		
	/** standard constructor */
	Proof(ODESystem system, double loKSI, double upKSI, double a_S, double b_S, double a_U, double b_U, double section = 3.0){
		this->loKSI = loKSI;
		this->upKSI = upKSI;
		this->system = system;
		this->a_S = a_S;
		this->b_S = b_S;
		this->a_U = a_U;
		this->b_U = b_U;
		this->section = section;
		
		//tell ode that we are computing using all posiible values of xi
		interval ksi = interval(loKSI, upKSI);
		this->system.setKSI(ksi);		
		IntervalVector singular2 = this->system.getSingular2();
		interval X2 = singular2[0], Y2 = singular2[1];
		//compute origin of all computations - z_0
		this->z0_x = X2.mid().leftBound();
		this->z0_y = Y2.mid().leftBound();
		info << "Point z_0 = (" << z0_x << ", " << z0_y << ")\n";
			
		//compute forbidden region
		this->forbiddenX = singular2[0];
		
		//compute interval matrix of eigenvectors for all possible stationary points and xi
		IntervalMatrix Df00 = this->system.getDerivAt(X2, Y2);		debug << "Df00 = \n" << Df00 << "\n";		
		P = i_diagonalize(Df00);  	debug << "P = \n" << P << "\n";			
		IP = inverse2x2(P);	 		debug << "IP o Df00 o P = \n" << IP * Df00 * P << "\n";		
		
		//compute coordinate change that we will use for all the sets
		Df00 = this->system.getDerivAt(z0_x, z0_y);		debug << "Df00 = \n" << Df00 << "\n";		
		IntervalMatrix M = diagonalize(Df00);
		//use only one of the matrices
		double c11, c12, c21, c22;
		c11 = M(1,1).mid().leftBound(); c12 = M(1,2).mid().leftBound();
		c21 = M(2,1).mid().leftBound(); c22 = M(2,2).mid().leftBound();
		M(1,1) = c11; M(1,2) = c12;
		M(2,1) = c21; M(2,2) = c22;		
		IntervalMatrix IM = inverse2x2(M);		debug << "M = \n" << M << "\n";	 debug << "IM o Df00 o M = \n" << IM * Df00 * M << "\n";		
		
		//do not change coordinates for unstable direction
		//this is better for forward time integration, as this allows 
		//stable direction to shrink size of the set, while the set
		//moves with more-or-less the same speed in unstable direction in
		//every point of the set. 
		//if we change coordinates, then one part of the set will move
		//faster than the other ant it will cause overgrowth of the set.
		M_U = M;
		info << "Matrix M_U for unstable set U = " << M << "\n";
		//for matrix M_S we make orthogonal matrix
		//because it is better for testing Isolating Block
		//(because coordinates of this system given by M
		// usually tilt the set S more, creating more twisted set)
		M_S = IntervalMatrix::Identity(2);
		M_S(1, 1) = -c22; M_S(1, 2) = c12;
		M_S(2, 1) =  c12; M_S(2, 2) = c22;
		info << "Matrix M_S for stable set S = " << M_S << "\n";
		
		//setup graphics:		
		//set xi to the whole interval \Xi				
		graphics.setup(0, z0_x - 4.0, z0_y - 1.0, z0_x + 1.0, z0_y + 6.0, z0_x, z0_y, section);			
		graphics.setup(1, z0_x - 1.5*max(a_S, a_U), z0_y - 1.5*max(b_S, b_U), z0_x + 1.5*max(a_S, a_U), z0_y + 1.5*max(b_S, b_U), z0_x, z0_y, section);			
	}	
	/** 
	 * rigorous computation of trajectory of estimates of stable and unstable sets until some section line
	 * in the end we verifies that for \xi_{lo} and \xi_{up} the images on section behave apropriate.
	 *
	 * @notice here we use krak for visualization.
	 */
	bool testExistence(double step, int order){
		info << "Starting Existence test\n";								
		//this is control variable that test whatever any exception during computation occured
		bool noException = true;
		
		//cone condition stable result - Df(U)^TQ + QDf(U) positive definite
		bool ccsResult = testConeCond(M_S, a_S, b_S);
		info << "testExistence(): Cone Condition for stable [" << (ccsResult ? "PASSED" : "FAILED") << "]\n";

		//cone condition unstable result - Df(U)^TQ + QDf(U) positive definite
		bool ccuResult = testConeCond(M_U, a_U, b_U);
		info << "testExistence(): Cone Condition for unstable [" << (ccuResult ? "PASSED" : "FAILED") << "]\n";	
		
		//make iteration sets 
		//(choose boundary fragments that lie left of singular points)
		Rect2Set prototypeU = ConeCondSet::makeUnstableSet(z0_x, z0_y, M_U, a_U, b_U, forbiddenX); graphics << Color::red << 1 << prototypeU << "Estimate on unstable exit point z_u(\\Xi) (red)\n" << 0 << prototypeU;
		Rect2Set prototypeS = ConeCondSet::makeStableSet(z0_x, z0_y, M_S, a_S, b_S, forbiddenX); graphics << Color::green << 1 << prototypeS << "Estimate on stable exit point z_s(\\Xi) (green)\n" << 0 << prototypeU;				
				
		IntervalVector RUI(prototypeU);	debug << "testExistence(): RUI = " << RUI <<"\n";
		IntervalVector RSI(prototypeS); debug << "testExistence(): RSI = " << RSI <<"\n";
		//this is temporary variable, used only to 'feed' intersection routine
		interval res = 0;				debug << "testExistence(): forbidden X = " << forbiddenX << "\n";		
		
		//test if singularites have no intersection with both iteration sets
		bool notSingularU = !intersection(RUI[0], forbiddenX, res) && !intersection(RUI[0], -forbiddenX, res) && !intersection((z0_x - (RUI[0] - z0_x)), forbiddenX, res) && !intersection((z0_x - (RUI[0] - z0_x)), -forbiddenX, res);
		info << "testExistence(): X(U) = " << RUI[0] << " u " << (z0_x - (RUI[0] - z0_x)) << ", X(Delta) = " << forbiddenX << " u " << (-forbiddenX) << "\n";
		info << "testExistence(): Unstable not singular? " << (notSingularU ? "PASSED" : "FAILED") << "\n"; 
		bool notSingularS = !intersection(RSI[0], forbiddenX, res) && !intersection(RSI[0], -forbiddenX, res) && !intersection((z0_x - (RSI[0] - z0_x)), forbiddenX, res) && !intersection((z0_x - (RSI[0] - z0_x)), -forbiddenX, res);
		info << "testExistence(): X(S) = " << RSI[0] << " u " << (z0_x - (RSI[0] - z0_x)) << ", X(Delta) = " << forbiddenX << " u " << (-forbiddenX) << "\n";
		info << "testExistence(): Stable not singular? " << (notSingularS ? "PASSED" : "FAILED") << "\n";																									
		
		//now we have our sets and they do not cross lines of singular points		
		
		//start iteration for $\xi=\xi_{lo}$
		info << "testExistence(): Integration for xi_{lo} = " << loKSI << "\n";			
		//start sets
		Rect2Set RU = prototypeU;
		Rect2Set RS = prototypeS;
		//out sets
		IntervalVector RUres, RSres;
		//start integration
		try{
			integration(RS, RU, RSres, RUres, loKSI, step, order);	
		}catch(...){
			noException = false;
		}		
		//setup last window - close up on the section
		double graphX = ((RUres[0] + RSres[0]) / 2.0).leftBound();
		double graphY = ((RUres[1] + RSres[1]) / 2.0).leftBound();
		//TODO: wyliczac te wartosci 0.005... 
		graphics.setup(2, graphX - 0.005, graphY - 0.005, graphX + 0.005, graphY + 0.005, z0_x, z0_y, section);			
		
		//output the sets on the section
		graphics << 2 << "Close-up on the section L\n";
		graphics << 2 << Color::red << RUres;   debug << "testExistence(): RU = " << IntervalVector(RU) << " RU res = " << RUres << "\n";
		graphics << 2 << Color::green << RSres;	debug << "testExistence(): RS = " << IntervalVector(RS) << " RS res = " << RSres << "\n";				
		
		//remember the bounds for x coordinate on the section
		interval piLoRU = IntervalVector(RUres)[0]; 
		info << "testExistence(): Image of U set on section = " << piLoRU << "\n";										
		interval piLoRS = IntervalVector(RSres)[0]; 
		info << "testExistence(): Image of S set on section = " << piLoRS << "\n";
	
		
		//start iteration for $\xi=\xi_{up}$
		info << "testExistence(): Integration for xi_{up} = " << upKSI << "\n";	
		//recreate sets
		RU = prototypeU;
		RS = prototypeS;
		//start integration
		try{
			integration(RS, RU, RSres, RUres, upKSI, step, order);	
		}catch(...){
			noException = false;
		}
		//output the sets on the section		
		graphics << 2 << Color::pine << RSres; debug << "testExistence(): RS = " << IntervalVector(RS) << " RS res = " << RSres << "\n";
		graphics << 2 << Color::red << RUres;  debug << "testExistence(): RU = " << IntervalVector(RU) << " RU res = " << RUres << "\n";
		
		interval piUpRU = IntervalVector(RUres)[0]; 
		info << "testExistence(): Image of U set on section = " << piUpRU << "\n";
		interval piUpRS = IntervalVector(RSres)[0]; 
		info << "testExistence(): Image of S set on section = " << piUpRS << "\n";
					

		//we have ale necessary components - do the final check
		bool result = ccsResult && ccuResult && noException && notSingularU && notSingularS && (((piLoRU < piLoRS) && (piUpRU > piUpRS)) || ((piLoRU > piLoRS) && (piUpRU < piUpRS)));
		if (result){
			info << "testExistence(): [PASSED] Images of sets U(xi) and S(xi) exchange positions on section!\n";
		}else{
			if (!(ccsResult && ccuResult)) info << "testExistence(): [FAILED] Sets U and S have not passed Isolating Block or Cone Condition Test...\n";
			else info << "testExistence(): [FAILED] Images of sets U(xi) and S(xi) intersects on section or not exchange positions...\n";
		}
		
		//this will stop proof and wait for keypress
		graphics << 0 << Keypress("EXISTENCE DONE, please press a key\n"); 					
		return result;
	}
	
	/** 
	 * rigorous computation of trajectory with estimates of stable and unstable sets until some section line 
	 * and computation of derivative of the flow on this section. We use here extended system (ie. with addition of $\xi' = 0$)
	 *
	 * @notice here we use krak for visualization.
	 */
	bool testUniqueness(double step, int order){
		info << "Starting Uniqueness test\n";						
		bool noException = true;
		
		//now we need to computer for all possible values of ksi, so we use ksi = $[\xi_lo, \xi_up]$
		interval ksi = interval(loKSI, upKSI);
		system.setKSI(ksi);	
		
		//Q = quadratic form for cone conditions 
		IntervalMatrix Q = IntervalMatrix::Identity(DIMENSION);
		Q(1,1) =  1; 
		Q(2,2) = -1;
		info << "testUniqueness(): Using Q = " << Q << "\n";
			
		//make iteration sets (choose boundary fragments)
		//suffix Param means 'with param', so its produce sets with additional dimension for $[\xi_lo, \xi_up]$		
		C1Rect2SetParam RU = ConeCondSet::makeUnstableSetParam(z0_x, z0_y, M_U, a_U, b_U, loKSI, upKSI, forbiddenX); RU.setOutput(graphics); debug << "testUniqueness(): RUI = " << IntervalVectorParam(RU) <<"\n";
		C1Rect2SetParam RS = ConeCondSet::makeStableSetParam(z0_x, z0_y, M_S, a_S, b_S, loKSI, upKSI, forbiddenX); RS.setOutput(graphics); debug << "testUniqueness(): RSI = " << IntervalVectorParam(RS) <<"\n";					
		
		//integrate system
		MapParam f = system.getMapParam();			
		//work in oryginal coordinates
		system.setTranslateParams(f, 0.0, 0.0);
		system.setTransformParams(f, IntervalMatrix::Identity(2));			
		
		FunctionParam s = "var:x,y,z;fun:y-3.0;"; //section
		TaylorParam TU(f, order,  step);
		//step control sometimes gives realy bad estimates, so we can turn it off
		if (turnOffStepControlForUnstable) TU.turnOffStepControl();
		TaylorParam TS(f, order, -step);
		//step control sometimes gives realy bad estimates, so we can turn it off
		if (turnOffStepControlForStable) TS.turnOffStepControl();

		PoincareMapParam pmu(TU,s,capd::poincare::MinusPlus);
		//step control sometimes gives realy bad estimates, so we can turn it off
		if (turnOffStepControlForUnstable) pmu.turnOffStepControl();		
		PoincareMapParam pms(TS,s,capd::poincare::PlusMinus);
		//step control sometimes gives realy bad estimates, so we can turn it off
		if (turnOffStepControlForStable) pms.turnOffStepControl();

		//here we will store derivatives of the flow
		IntervalMatrixParam base_derivS;
		IntervalMatrixParam base_derivU;
		
		//here we will store maps of starting sets
		IntervalVectorParam RUVres;		
		IntervalVectorParam RSVres;				
		
		//start integration
		try{
			graphics << 0 << Color::orange;
			RUVres = pmu(RU, base_derivU);
		} catch (std::exception &e){
			info << "testUniqueness(): [FAILED] Exception in PMU: \n" << e.what() << "\n";
			noException = false;
			graphics << Color::orange << IntervalVectorParam(RU);
		}
		//here we need to execute correction routine that will return true derivatives of the poincare map
		IntervalVectorParam dT_U = pmu.computeDT(RUVres, base_derivU);
		info << "base deriv U = " << base_derivU << "\n";
		info << "dT_U = " << dT_U << "\n";
		IntervalMatrixParam derivU = pmu.computeDP(RUVres, base_derivU, dT_U);
		try{
			graphics << 0 << Color::blue;
			RSVres = pms(RS, base_derivS);
		} catch (std::exception &e){
			info << "testUniqueness(): [FAILED] Exception in PMS: \n" << e.what() << "\n";
			noException = false;
			graphics << Color::blue << IntervalVectorParam(RS);
		}				
		//here we need to execute correction routine that will return true derivatives of the poincare map
		IntervalVectorParam dT_S = pms.computeDT(RSVres, base_derivS);
		info << "base deriv S = " << base_derivS << "\n";
		info << "dT_S = " << dT_S << "\n";
		IntervalMatrixParam derivS = pms.computeDP(RSVres, base_derivS, dT_S);		
		
		//here we have Dm matrix of m w.r.t initial conditions
		info << "testUniqueness(): Dm on S: \n" << derivS << "\n"; debug << "testUniqueness(): RSV res = " << RSVres <<"\n";
		info << "testUniqueness(): Dm on U: \n" << derivU << "\n"; debug << "testUniqueness(): RUV res = " << RUVres <<"\n";
		
		//now we need to compute derivative of the map w.r.t. $\xi$ (let Z be the function mapping (un)stable manifold onto section).
		//dP/dKSI = deltaP/deltaKSI + deltaP/deltaX * a * deltaZ/deltaKSI + deltaP/deltaY * c * deltaZ / deltaKSI
		//of course we have only estimates on Z, P, etc.
		
		//now we need estimates on the derivatives of initial conditions w.r.t xi		
		//Compute L constants
		interval cL_S = getL(M_S, a_S, b_S);
		interval cL_U = getL(M_U, a_U, b_U);		
		
		interval cA_S = getA(M_S, Q, a_S, b_S);
		interval cA_U = getA(M_U, Q, a_U, b_U); 	
		
		//i_sqrt_delta = 1/sqrt(delta) = 2 * L / A, we take bigger 1/sqrt(delta) because normally we should take lower delta.		
		double delta_S = power((interval(cA_S) / (interval(2.0) * interval(cL_S))), 2).leftBound(); 
		double delta_U = power((interval(cA_U) / (interval(2.0) * interval(cL_U))), 2).leftBound();
		
		double i_sqrt_delta_S = (interval(2.0) * (interval(cL_S) / interval(cA_S))).rightBound(); 
		double i_sqrt_delta_U = (interval(2.0) * (interval(cL_U) / interval(cA_U))).rightBound();		
		
		info << "testUniqueness(): Stable manifold estimates: \n";
		info << "testUniqueness(): A = " << cA_S << ", L = " << cL_S << ", delta <= " << delta_S << ", 1/sqrt(delta) = " << i_sqrt_delta_S << "\n";
		info << "testUniqueness(): Unstable manifold estimates: \n";
		info << "testUniqueness(): A = " << cA_U << ", L = " << cL_U << ", delta <= " << delta_U << ", 1/sqrt(delta) = " << i_sqrt_delta_U << "\n";
		
		interval dPdzS = interval(-i_sqrt_delta_S, i_sqrt_delta_S); debug << "dPdzS = " << dPdzS << " ( " << a_S << ", " << a_S * dPdzS << " )\n";
		interval dPdzU = interval(-i_sqrt_delta_U, i_sqrt_delta_U); debug << "dPdzU = " << dPdzU << " ( " << b_U << ", " << b_U * dPdzU << " )\n";		
			
		info << "testUniqueness(): z_s'(xi) = (" << M_S(1,1) * a_S * dPdzS << ", " << M_S(1,2) * a_S * dPdzS << ")^T\n";
		info << "testUniqueness(): z_u'(xi) = (" << M_U(1,2) * b_U * dPdzU << ", " << M_U(2,2) * b_U * dPdzU << ")^T\n";
			
		interval estimateS = derivS(1,3) + derivS(1,1) * M_S(1,1) * a_S /* a */ * dPdzS + derivS(1,2) * M_S(1,2) * a_S /* c */ * dPdzS;
		interval estimateU = derivU(1,3) + derivU(1,1) * M_U(1,2) * b_U /* b */ * dPdzU + derivU(1,2) * M_U(2,2) * b_U /* d */ * dPdzU;
		interval estimate = estimateS - estimateU;
		
		info << "testUniqueness(): estimate = " << estimate << " (estimate = estimate on S - etimate on U)\n";
		info << "testUniqueness(): " << " estimate on U = " << estimateU << " estimate on S " << estimateS << "\n";
		
		if (noException && (estimate < 0 || estimate > 0)){
			info << "testUniqueness(): [PASSED] Estimate of difference of derivatives is strictly positive or negative guaranteing uniqueness.\n";
		}else{
			info << "testUniqueness(): [FAILED] Cannot find estimates guaranteing uniqueness...\n";
		}
		
		graphics << 0 << Keypress("UNIQUENES TEST DONE, please press a key\n");
		
		return noException && (estimate < 0 || estimate > 0);
	}
	/**
	 * Test for finite time approach of solution to a statonary point.
	 */
	bool testFiniteTime(){		
		info << "Starting Finite Time test\n";	
		
		info << "V_1 = (" << P(1,1) << ", " << P(2,1) << ")^T\n";	
		info << "V_2 = (" << P(1,2) << ", " << P(2,2) << ")^T\n";	
		
		bool result = !(P(1,1).contains(0.0) || P(1,2).contains(0.0));
		
		info << "testFiniteTime(): Eigenvectors has nonzero first coordinate? [" << (result ? "PASSED" : "FAILED") << "]\n"; 
				
		//test if eigenvectors have first coordinate not containing 0.0
		return result;
	}

};
/* class Proof *******************************************************************/

/*****************************************************************************
 * MAIN simple as it is
 *****************************************************************************/
int main(int, char *[]){
	//set precision of the output numbers
	cout.precision(10);
	//create ODE's
	ODESystem system(KAPPA, TAU, GAMMA, CHI);
	//get Critical value of KSI for which periodic orbit exits (calculated from V. Vladimirov)
	double KSICr = system.getCriticalKSI();	
	//set up test interval in which we look up KSI_cr_2, value for which homoclinic loop exists
	double loKSI = 0.0, upKSI = 0.0;
	if (loKSIMode) loKSI += (XI + interval(KSICr)).leftBound(); else loKSI = XI.leftBound();
	if (upKSIMode) upKSI += (XI + interval(KSICr)).rightBound(); else upKSI = XI.rightBound();
	
	std::cout << "Homoclinic existence check for:\n";
	std::cout << "a_U = " << a_U << ", b_U = " << b_U << "\n";	
	std::cout << "a_S = " << a_S << ", b_S = " << b_S << "\n";	
	std::cout << "Xi = [" << loKSI << ", " << upKSI << "] (Xi_cr = " << KSICr << ")\n";	
	std::cout << "\n";
	
	Proof<CoutOutput, DebugOutput, GraphicsOutput> proof(system, loKSI, upKSI, a_S, b_S, a_U, b_U);	
	
	//homoclinic orbit result - continuation of exit points from sets U and S
	bool exsResult = proof.testExistence(step, order);	
	std::cout << "=== Existence test [" << (exsResult ? "PASSED" : "FAILED") << "] === " << std::endl;
	
	//uniqueness test - derivative estimates for map T : \Xi -> \R^2
	bool unqResult = proof.testUniqueness(step, order); 	
	std::cout << "=== Uniqueness test [" << (unqResult ? "PASSED" : "FAILED") << "] === " << std::endl;
	
	//test finite time by checking x coordinate of eigenvectors
	bool ftResult = proof.testFiniteTime();	
	std::cout << "=== Finite Time test [" << (ftResult ? "PASSED" : "FAILED") << "] === " << std::endl;	
	
	std::cout << "\nResult of the proof: \n";
	if (exsResult){
		std::cout << "[SUCCESS] Homoclinic orbit exists";
		if (unqResult){
			std::cout << " AND is unique"; 
		}
		if (ftResult){
			std::cout << " AND the time is finite"; 
		}
		std::cout << "!";
	}else{
		std::cout << "[FAILED] We have nothing...";
	}

	return 0;
}
/* end of main **********************************************************************************/






/************************************************************************************************
 * helper functions of the proof ****************************************************************
 ************************************************************************************************/
template<typename InfoHandler, typename DebugHandler, typename GraphicsHandler> 
IntervalMatrix Proof<InfoHandler, DebugHandler, GraphicsHandler>::coneCondMatrix(IntervalMatrix M, IntervalMatrix Q, double a, double b){
	//do computation for the whole interval \Xi
	interval ksi = interval(loKSI, upKSI);
	system.setKSI(ksi);
	
	Function deriv[2][2];
	//get the partial derivatives of the system
	system.getDeriv(deriv);
	Function& df1dx = deriv[0][0];
	Function& df1dy = deriv[0][1];
	Function& df2dx = deriv[1][0];
	Function& df2dy = deriv[1][1];	
	
	//move orgin to the point z_0 (Df o z_0)
	system.setTranslateParams(df1dx, z0_x, z0_y);
	system.setTranslateParams(df1dy, z0_x, z0_y);
	system.setTranslateParams(df2dx, z0_x, z0_y);
	system.setTranslateParams(df2dy, z0_x, z0_y);			
	
	//make real C - coordinate change for f_c - it's VERY important for the assuptions of the theorems!
	IntervalMatrix s = IntervalMatrix::Identity(2);
	s(1,1) = a;
	s(2,2) = b;	
	IntervalMatrix IC = M * s; //because c^-1_N = M * s	
	IntervalMatrix C = inverse2x2(IC);	
	
	//made it Df o z_0 o c_N
	system.setTransformParams(df1dx, IC);
	system.setTransformParams(df1dy, IC);
	system.setTransformParams(df2dx, IC);
	system.setTransformParams(df2dy, IC);
	
	//do the test
	//because we set translate params to C_n and move orgin to z0 we can 'feed' 
	//our computations with simple intervals
	interval x(-1.0, 1.0);
	interval y(-1.0, 1.0);
	
	df1dx.setParameter("y", y);
	df1dy.setParameter("y", y);
	df2dx.setParameter("y", y);
	df2dy.setParameter("y", y);	
	
	//We work in good coordinates Dc_N o f o c^-1_N = f_c
	//we have: Dc_N = variable C
	//         c_N^-1 = IC + setTranslateParams() 
	//DCofoIC = D (Dc_N o f o c^-1_N) = C * Df(IC) * IC
	IntervalMatrix DCofoIC = IntervalMatrix::Identity(2);
	DCofoIC(1,1) = df1dx(x);
	DCofoIC(1,2) = df1dy(x);
	DCofoIC(2,1) = df2dx(x);
	DCofoIC(2,2) = df2dy(x);
	DCofoIC = C * DCofoIC * IC;	
	
	//this is exactly cone condition matrix in good cordinates c_N
	//since we have c_N = const matrix
	IntervalMatrix DfQQDf = Transpose(DCofoIC)*Q+Q*(DCofoIC);
	
	return DfQQDf;
}

template<typename InfoHandler, typename DebugHandler, typename GraphicsHandler> 
bool Proof<InfoHandler, DebugHandler, GraphicsHandler>::testConeCond(IntervalMatrix M, double a, double b){		
	info << "testConeCond(): start\n";
		
	interval ksi = interval(loKSI, upKSI);
	system.setKSI(ksi);
	
	//test Isolating Block
	Map F = system.getMap();
	system.setTranslateParams(F, z0_x, z0_y);
	system.setTransformParams(F, M);
	debug << "testConeCond(): Isolating Block Test:\n";
	//we do this by cutting intervals into small pieces and compute inner product 
	//of the vector field on it with the normal to the given boundary
	//d is arbitrary constant... TODO: THINK: should it be defined in input?
	int d = 4096;
	int badv = 0;
	int badd = 0;
	interval globdotv;
	interval globdotw;
	for (int i = -d; i < d; i++){
		//on the unstable exit set
		IntervalVector v1 = F(makeVector(interval(a), interval((b / d) * i, (b / d) * (i+1)) ));
		IntervalVector n1 = M * makeVector(1.0, 0.0);
		IntervalVector v2 = F(makeVector(interval(-a), interval((b / d) * i, (b / d) * (i+1)) ));
		IntervalVector n2 = M * makeVector(-1.0, 0.0);
		//on the stable exit set
		IntervalVector w1 = F(makeVector(interval((a / d) * i, (a / d) * (i+1)), interval(b)));
		IntervalVector k1 = M * makeVector(0.0, 1.0);
		IntervalVector w2 = F(makeVector(interval((a / d) * i, (a / d) * (i+1)), interval(-b)));
		IntervalVector k2 = M * makeVector(0.0, -1.0);

		//test if any of the vector sets spans whole space (contains (0,0)) - this is realy bad... 
		if (v1[0].contains(0.0) && v1[1].contains(0.0)){badv++; debug << "testConeCond(): BAD v1 = " << v1 << " for i = " << i << "\n";}
		if (v2[0].contains(0.0) && v2[1].contains(0.0)){badv++; debug << "testConeCond(): BAD v2 = " << v2 << " for i = " << i << "\n";}
		if (w1[0].contains(0.0) && w1[1].contains(0.0)){badv++; debug << "testConeCond(): BAD w1 = " << w1 << " for i = " << i << "\n";}
		if (w2[0].contains(0.0) && w2[1].contains(0.0)){badv++; debug << "testConeCond(): BAD w2 = " << w2 << " for i = " << i << "\n";}
		//compute inner products to test if the vector field points in good direction on each set
		if (!(dot(v1, n1) > 0.0)){ badd++; debug << "testConeCond(): BAD dot(v1, n1) = " << dot(v1, n1) << " for i = " << i << "\n";}
		if (!(dot(v2, n2) > 0.0)){ badd++; debug << "testConeCond(): BAD dot(v2, n2) = " << dot(v2, n2) << " for i = " << i << "\n";}
		if (!(dot(w1, k1) < 0.0)){ badd++; debug << "testConeCond(): BAD dot(w1, k1) = " << dot(w1, k1) << " for i = " << i << "\n";}
		if (!(dot(w2, k2) < 0.0)){ badd++; debug << "testConeCond(): BAD dot(w2, k2) = " << dot(w2, k2) << " for i = " << i << "\n";}
		
		if (i == -d){
			globdotv = capd::intervals::intervalHull(dot(v1, n1), dot(v2, n2));
			globdotw = capd::intervals::intervalHull(dot(w1, k1), dot(w2, k2));
		}else{
			globdotv = capd::intervals::intervalHull(globdotv, dot(v1, n1));
			globdotv = capd::intervals::intervalHull(globdotv, dot(v2, n2));
			globdotw = capd::intervals::intervalHull(globdotw, dot(w1, k1));
			globdotw = capd::intervals::intervalHull(globdotw, dot(w2, k2));
		}
	}			
	info << "testConeCond(): Inner product on N- = " << globdotv << " (should be positive)\n";
	info << "testConeCond(): Inner product on N+ = " << globdotw << " (should be negative)\n";
	if (badd == 0){		
		info << "testConeCond(): [SUCCESS] The set is Isolating Block!\n";
	} else {
		info << "testConeCond(): [FAIL] Isolating Block Test failed in " << badd << " intervals...\n";
	}
	
	//Q = quadratic form for cone conditions 
	IntervalMatrix Q = IntervalMatrix::Identity(DIMENSION);
	Q(1,1) =  1; 
	Q(2,2) = -1;
	info << "testConeCond(): Using Q = " << Q << "\n";
	
	IntervalMatrix DfQQDf = coneCondMatrix(M, Q, a, b);
	
	info << "testConeCond(): Test matrix (Df_c^T)Q+Q(Df_c) is \n" << DfQQDf << "\n\n";
	
	interval m1 = DfQQDf(1,1);
	info << "testConeCond(): |minor 1| = " << m1 << "\n";
	interval m2 = DfQQDf(1,1)*DfQQDf(2,2) - DfQQDf(1,2) * DfQQDf(2,1);
	info << "testConeCond(): |minor 2| = " << m2 << "\n";	
	
	if (m1 > 0 && m2 > 0){
		info << "testConeCond(): [SUCCESS] Cone Condition Matrix is positive definite!\n";
	} else {
		info << "testConeCond(): [FAIL] Cannot assure that matrix is positive definite...\n";
	}
	
	info << "testConeCond(): finished\n";
	return (badd == 0) && m1 > 0 && m2 > 0;
}


template<typename InfoHandler, typename DebugHandler, typename GraphicsHandler> 
void Proof<InfoHandler, DebugHandler, GraphicsHandler>::integration(Rect2Set& S, Rect2Set& U, IntervalVector& outS,IntervalVector& outU, double ksi, double step, int order){
	info << "integration(): start for xi = " << ksi << "\n";
	system.setKSI(ksi);		
	Map f = system.getMap();
	
	//set parameters for iteration
	system.setTranslateParams(f, 0, 0);
	system.setTransformParams(f, IntervalMatrix::Identity(2));											
	
	Function s = "var:x,y,z;fun:y-3.0;"; //section
	s.setParameter("s", section);
	Taylor TU(f, order,  step); //taylor method for unstable manifold
	Taylor TS(f, order, -step); //taylor method for stable manifold
	
	PoincareMap pmu(TU, s, capd::poincare::MinusPlus);
	 //step control sometimes gives realy bad estimates, so we can turn it off
	if (turnOffStepControlForUnstable) pmu.turnOffStepControl();
	PoincareMap pms(TS, s, capd::poincare::PlusMinus);
	 //step control sometimes gives realy bad estimates, so we can turn it off
	if (turnOffStepControlForStable) pms.turnOffStepControl();
	
	graphics << 0 << Color::red << U;
	graphics << 0 << Color::green << S;
	
	U.setOutput(graphics);
	S.setOutput(graphics);
	
	//results on section (map of sets RU and RS on the section s)
	IntervalVector RUres, RSres;
	try {
		graphics << Color::red;
		outU = pmu(U);
	}catch (std::exception &e){
		info << "integration(): [FAIL] Exception in Poincare Map for U: " << e.what() << "\n";
		throw;
	}
	try {
		graphics << Color::green;
		outS = pms(S);
	}catch (std::exception &e){			
		info << "integration(): [FAIL] Exception in Poincare Map for S: " << e.what() << "\n";
		throw;
	}
	info << "integration(): finished\n";
}

template<typename InfoHandler, typename DebugHandler, typename GraphicsHandler> 
interval Proof<InfoHandler, DebugHandler, GraphicsHandler>::getL(IntervalMatrix M, double a, double b){
	Function ds[2];
	system.getDerivParam(ds);
	
	//make real C - coordinate change for f_c -its VERY important for the assuptions of the theorems!
	IntervalMatrix s = IntervalMatrix::Identity(2);
	s(1,1) = a;
	s(2,2) = b;	
	IntervalMatrix IC = M * s; //because c^-1_N = M * s
	IntervalMatrix C = inverse2x2(IC);
	
	system.setTranslateParams(ds[0], z0_x, z0_y);
	system.setTransformParams(ds[0], IC);
	system.setTranslateParams(ds[1], z0_x, z0_y);
	system.setTransformParams(ds[1], IC);
			
	interval x(-1.0, 1.0), y(-1.0, 1.0);
	ds[0].setParameter("y", y);
	ds[1].setParameter("y", y);
	
	interval X1 = ds[0](x);
	interval X2 = ds[1](x);
	debug << "X1 = " << X1 << ", X2 = " << X2 << "\n";
	interval X1_c = C(1,1) * X1 + C(1,2) * X2; 
	interval X2_c = C(2,1) * X1 + C(2,2) * X2;
	//we need to take upper bound because (1/delta) = sqrt(4 * L^2 / A^2)
	return capd::intervals::sqrt((capd::intervals::power(X1_c, 2) + capd::intervals::power(X2_c, 2)));	
}

template<typename InfoHandler, typename DebugHandler, typename GraphicsHandler> 
interval Proof<InfoHandler, DebugHandler, GraphicsHandler>::getA(IntervalMatrix M, IntervalMatrix Q, double a, double b){	
	IntervalMatrix DfQQDf = coneCondMatrix(M, Q, a, b);	
	//we use binsearch to find good value A for S (good_guess_S)
	//take lower value - thus always first minor will be positive
	interval lambda_up = DfQQDf(1,1) > DfQQDf(2,2) ? DfQQDf(2,2) : DfQQDf(1,1);
	//in this range we search for good A
	lambda_up = interval(0.0, lambda_up.rightBound());
	double good_guess = 0.0;
	double guess;
	while (std::abs(lambda_up.rightBound() - lambda_up.leftBound()) > 0.00001){			
		//get the middle
		guess = lambda_up.mid().leftBound();
		if (((DfQQDf(1,1) - guess) * (DfQQDf(2,2) - guess)) - (DfQQDf(1,2)*DfQQDf(2,1)) > 0){
			//OK - positive definite
			good_guess = guess;
			lambda_up = interval(guess, lambda_up.rightBound());
		}else{
			//bad - not positive definite
			lambda_up = interval(lambda_up.leftBound(), guess);
		}
	}
	return interval(good_guess);
}
